//------------------------------------------------------------------------------

#include "CommonBase.h"
#include "PMsg_p.hpp"
#include "Pglobals.h"
#include <stdio.h>
#include <algorithm>

#include <string>
using namespace std;
typedef unsigned char byte;
const string CommonBase::S00 = string();
const char* CommonBase::MPISvc = "POETS_MPI_Master";

//==============================================================================

CommonBase::CommonBase(int argc,char ** argv,string d,string src)
{
//MPI_Init(&argc,&argv);               // Start up MPI
int MPI_provided;
MPI_Init_thread(&argc,&argv,MPI_THREAD_MULTIPLE,&MPI_provided);

AcceptConns = false;
const int SNDBUFSIZ = 1000000000;      // MPI immediate message send buffer size
char * SNDBUF = (char *)malloc(SNDBUFSIZ);             // Pull it off the heap
if (SNDBUF!=0) MPI_Buffer_attach(SNDBUF,SNDBUFSIZ);    // Attach it to MPI
Msgbufsz = 2 << LOG_MSGBUF_BLK_SZ;
MPI_Buf = new char[Msgbufsz];            // and set up a receive buffer
Comms.push_back(MPI_COMM_WORLD);         // Set up the default comm
pPmap.push_back(new ProcMap(this));      // Create default universe process map
int lUsize;
MPI_Comm_size(Comms[0], &lUsize);
Usize.push_back(lUsize);                // record the size of the local universe

Sderived = d;                          // Derived class name

FnMapx.push_back(new FnMap_t); // create a default function table (for the *base* class!)
// Load event handler map
// the 'SystConn'event, instructing processes to participate
// in setup of a connection to another group of processes, only applies to
// the local communicator. One cannot do this over an intercommunicator, so
// these functions are only registered for the local function table.
(*FnMapx[0])[PMsg_p::KEY(Q::EXIT                )] = &CommonBase::OnExit;
(*FnMapx[0])[PMsg_p::KEY(Q::PMAP                )] = &CommonBase::OnPmap;
(*FnMapx[0])[PMsg_p::KEY(Q::SYST,Q::CONN        )] = &CommonBase::OnSystConn;
(*FnMapx[0])[PMsg_p::KEY(Q::SYST,Q::ACPT        )] = &CommonBase::OnSystAcpt;

Prologue(argc,argv);                   // Start MPI, build process map.....

}

//------------------------------------------------------------------------------

CommonBase::~CommonBase()
{
//printf("********* CommonBase rank %d destructor\n",Urank); fflush(stdout);
WALKVECTOR(ProcMap*,pPmap,P)
  delete *P;
int idummy;                            // Clean up MPI immediate message buffer
void *SNDBUF;
WALKVECTOR(FnMap_t*,FnMapx,F)          // WALKVECTOR and WALKMAP are in macros.h (long include chain)
  delete *F;                           // get rid of function tables
// the seemingly daft argument to MPI_Comm_disconnect: &(*C) is correct here because
// MPI_Comm_disconnect can't work with an iterator.
for (vector<MPI_Comm>::iterator C = ++Comms.begin(); C != Comms.end(); C++) MPI_Comm_disconnect(&(*C)); // and comms other than MPI_COMM_WORLD
MPI_Buffer_detach(&SNDBUF,&idummy);    // (Blocks until exhausted)
free(SNDBUF);                          // Now we can kill it
delete [] MPI_Buf;                     // as well as the receive buffer
MPI_Finalize();
}

//------------------------------------------------------------------------------

void* CommonBase::Accept(void* par)
/* Blocking routine to connect to another MPI universe by publishing a port.
   This operates in a separate thread to avoid blocking the whole process.
*/
{
CommonBase* parent=static_cast<CommonBase*>(par);
// The leader of the intercomm will be the lowest-ranking Mothership, if there is one,
// or the zero-rank process otherwise.
MPI_Group MPI_Group_world, MPI_Group_remote;
MPI_Comm_group(parent->Comms[0],&MPI_Group_world);
while (parent->AcceptConns)
{
// run the blocking accept itself.
if (MPI_Comm_accept(parent->MPIPort,MPI_INFO_NULL,parent->Lrank,MPI_COMM_WORLD,&(parent->Tcomm)))
{
   // Might try Post() but failure could indicate an unreliable MPI universe
   // so all bets are off about whether any MPI communication would actually succeed.
   printf("Error: attempt to connect to another MPI universe failed\n");
   parent->AcceptConns = false;
   break;
}
int RemoteSize, RemoteRank; // detect if this is an exit by looking for a self-connection
int RootRank = 0;
MPI_Comm_remote_size(parent->Tcomm,&RemoteSize);
MPI_Comm_remote_group(parent->Tcomm,&MPI_Group_remote);
MPI_Group_translate_ranks(MPI_Group_world,1,&RootRank,MPI_Group_remote,&RemoteRank);
// printf("Connected to remote group of size %d, leader rank %d\n",RemoteSize,RemoteRank==MPI_UNDEFINED? -1 : RemoteRank);
// fflush(stdout);
if ((RemoteSize>1) || (RemoteRank == MPI_UNDEFINED)) // as long as we aren't shutting down,
{
   // trigger the Connect process in the main thread to complete the setup
   PMsg_p Creq;
   string N("");              // zero-length string indicates a server-side connection
   Creq.Put(2,&N);
   Creq.Key(Q::SYST,Q::CONN);
   Creq.Src(parent->Urank);
   Creq.comm = MPI_COMM_WORLD;
   Creq.Send(parent->Urank);
   while (parent->Tcomm != MPI_COMM_NULL); // block until connect has succeeded
}
else
{
   printf("Process rank %d received a shutdown connect message\n", parent->Urank);
   fflush(stdout);
   MPI_Comm_disconnect(&(parent->Tcomm)); // for completeness, though we are exiting.
   break;
}
}
pthread_exit(par);
return par;
}

//------------------------------------------------------------------------------

unsigned CommonBase::Connect(string svc)
// connects this process' MPI universe to a remote universe that has published
// a name to access it by.
{    
int error = MPI_SUCCESS;
// a server has its port already so can just open a comm
if (svc=="") Comms.push_back(Tcomm);
else // clients need to look up the service name
{
   MPI_Comm newcomm;
   char port[MPI_MAX_PORT_NAME];
   // Get the published port for the service name asked for.
   // Exit if we don't get a port, probably because the remote universe isn't
   // initialised yet (we can always retry).
   if (error = MPI_Lookup_name(svc.c_str(),MPI_INFO_NULL,port)) return error;
   // now try to establish the connection itself. Again, we can always retry.
   // if (error = MPI_Comm_connect(port,MPI_INFO_NULL,0,MPI_COMM_WORLD,&newcomm)) return error;
   if (error = MPI_Comm_connect(port,MPI_INFO_NULL,0,Comms[0],&newcomm)) return error;
   Comms.push_back(newcomm); // as long as we succeeded, add to the list of comms
}
int rUsize;
MPI_Comm_remote_size(Comms.back(), &rUsize);
Usize.push_back(rUsize);       // record the size of the remote universe
FnMapx.push_back(new FnMap_t); // give the new comm some function tables to use
pPmap.push_back(new ProcMap(this));  // and a new processor map for the remote group
PMsg_p prMsg;
SendPMap(Comms.back(), &prMsg);        // Send our process data to the remote group
int fIdx=FnMapx.size()-1;
// populate the new function table with the global functions
(*FnMapx[fIdx])[PMsg_p::KEY(Q::EXIT                )] = &CommonBase::OnExit;
(*FnMapx[fIdx])[PMsg_p::KEY(Q::PMAP                )] = &CommonBase::OnPmap;
if (svc=="") Tcomm = MPI_COMM_NULL;  // release any Accept comm.
return error;
}

//------------------------------------------------------------------------------

void CommonBase::MPISpinner()
// Spinner to handle ALL incoming MPI packets. There is only one exit from this:
// Decode() returns 1 when an "exit" command is detected.
{
for (;;) {
// See if there are any MPI packets coming down the pipe
  MPI_Status status;                   // Note the multi-threaded MPI probe
  int flag;
  for (unsigned cIdx = 0; cIdx < Comms.size(); cIdx++) // for all active comms
  {
    MPI_Iprobe(MPI_ANY_SOURCE,MPI_ANY_TAG,Comms[cIdx],&flag,&status);
    if (flag==0) {                       // Nothing there....
      // OnIdle();                          // Guess
      continue;                          // And try again
    }
    int count;
    MPI_Get_count(&status,MPI_CHAR,&count);
//  printf("MPISpinner: %d bytes coming\n",count);  fflush(stdout);
//  vector<char> buf(count+1);
    if (count > Msgbufsz) {             // Ensure we have the space for it
      delete [] MPI_Buf;                  // Throw the too-small one away
      Msgbufsz = ((count>>LOG_MSGBUF_BLK_SZ)+1)<<LOG_MSGBUF_BLK_SZ;
      MPI_Buf = new char[Msgbufsz]; // Pull a new (bigger) one off the heap
    }
    MPI_Request request;
    MPI_Irecv(&MPI_Buf[0],count,MPI_CHAR,MPI_ANY_SOURCE,MPI_ANY_TAG,Comms[cIdx],&request);
    int i=0;
    do { i++; MPI_Test(&request,&flag,&status); } while (flag==0);
//  printf("MPISpinner: after %d, %s has landed\n",i,&MPI_Buf[0]);  fflush(stdout);
    PMsg_p Pkt((unsigned char *)&MPI_Buf[0],count);// Turn it into a packet
    if (Decode(&Pkt,cIdx)!=0) return;    // Do it, possibly leaving afterwards
  }                                    
}
}

//------------------------------------------------------------------------------

unsigned CommonBase::OnExit(PMsg_p * Z,unsigned cIdx)
// Do not post anything further here - the LogServer may have already gone
{
WALKVECTOR(MPI_Comm,Comms,C) MPI_Barrier(*C);
// printf("(%s)::CommonBase::OnExit \n",Sderived.c_str());    fflush(stdout);
if (AcceptConns) // Accept thread running?
{
   printf("(%s)::CommonBase::OnExit closing down Accept MPI request on rank %d\n",Sderived.c_str(),Urank);
   fflush(stdout);
   // have to close the Accept thread via a matching MPI_Comm_connect
   // because the MPI interface has made no provision for a nonblocking accept, thus
   // otherwise the Accept thread will block forever waiting for a message that will
   // never come because we are shutting down. See Gropp, et al. "Using Advanced MPI"
   if (!Urank)
   {
   MPI_Comm dcomm;
   printf("Accept port name: %s\n", MPIPort);
   fflush(stdout);
   MPI_Comm_connect(MPIPort,MPI_INFO_NULL,0,MPI_COMM_SELF,&dcomm);
   }
   // printf("(%s)::CommonBase::OnExit waiting for Accept thread to exit\n",Sderived.c_str());
   // fflush(stdout);
   pthread_join(MPI_accept,NULL);
   AcceptConns = false; // stop accepting connections
   // printf("(%s)::CommonBase::OnExit shutting down\n",Sderived.c_str());
   // fflush(stdout);
}
if (Urank == Lrank)
{
   MPI_Unpublish_name(MPISvc,MPI_INFO_NULL,MPIPort);
   MPI_Close_port(MPIPort);
}
return 1;
}

//------------------------------------------------------------------------------

unsigned CommonBase::OnPmap(PMsg_p * Z,unsigned cIdx)
{
pPmap[cIdx]->Register(Z);                    // Load one element into the process map
if ((cIdx == 0) && (pPmap[0]->vPmap.size() == Usize[0])) // once we have everybody local
{
   Lrank = 0; // intercomm leader (lowest-rank mothership or 0)
   if (Urank == Lrank) // and if we are the leader,
   {
       MPI_Open_port(MPI_INFO_NULL,MPIPort); // Announce to MPI that we are open for business
       MPI_Publish_name(MPISvc,MPI_INFO_NULL,MPIPort); // and make us publicly available
       string port_name(MPIPort); // all local processes should get the port name

       // start up the Accept thread everywhere.
       PMsg_p Areq;
       Areq.Key(Q::SYST,Q::ACPT);
       Areq.Src(Urank);
       Areq.comm = Comms[0];
       Areq.Put(2,&port_name);
       Areq.Bcast();
       return OnSystAcpt(&Areq,0); // including ourselves.
   }
}
return 0;
}

//------------------------------------------------------------------------------

unsigned CommonBase::OnSystAcpt(PMsg_p * Z,unsigned cIdx)
{
if (cIdx) return 1; // Accept is only valid on the process' native comm.
string port_name;
Z->Get(2,port_name);
strcpy(MPIPort,port_name.c_str());
printf("Accept port name for process rank %d: %s\n", Urank, MPIPort);
fflush(stdout);
AcceptConns = true; // set the flag enabling accepts
void* args = this;                   // spin off a thread to accept MPI connections
if (pthread_create(&MPI_accept,NULL,Accept,args))  // from other universes
{
AcceptConns = false; // failure is fatal.
return 1;
}
return 0;
}

//------------------------------------------------------------------------------

unsigned CommonBase::OnSystConn(PMsg_p * Z,unsigned cIdx)
{
if (cIdx) return 1; // can only connect on local intracomm
string svc;
Z->Get(2,svc);              // What is our service name?
return Connect(svc);        // connect as requested.
}

//------------------------------------------------------------------------------

void CommonBase::Prologue(int argc,char ** argv)
// Startup code for EVERY orchestrator process.
{
int tUsize;
MPI_Comm_size(Comms[0],&tUsize); // Universe size
Usize.push_back(tUsize);
MPI_Comm_rank(Comms[0],&Urank);  // My place within it
char Uname[MPI_MAX_PROCESSOR_NAME];
MPI_Get_processor_name(Uname,&Ulen);   // Machine name
Sproc = string(Uname);                 // Translated from the original FORTRAN
strcpy(MPIPort,"PORT_NULL");           // Default MPI port
Tcomm = MPI_COMM_NULL;                 // initialise the Accept intercommunicator
 
PMsg_p Pkt;
SendPMap(Comms[0], &Pkt);        // Send our data to everybody else
pPmap[0]->Register(&Pkt);        // Need to put ourselves in seperately
MPI_Barrier(Comms[0]);           // Wait until everyone has told everyone
}

//------------------------------------------------------------------------------

void CommonBase::SendPMap(MPI_Comm comm, PMsg_p* Pkt)
{
// Shove the data all into a packet:
Pkt->Put(0,Urank);                      // Load my local rank...
Pkt->Put(1,&Sproc);                     // ... processor name
Pkt->Put(2,&Sderived);                  // ... C++ class
Pkt->Key(Q::PMAP);                      // Message key
Pkt->Src(Urank);
Pkt->comm = comm;                       // communicator over which this is being sent
                                        // Build process map
Pkt->Bcast();                           // via a broadcast
}

//==============================================================================


